"""Functional transform for Hessian vector product."""

from typing import Callable
from functools import partial

from .numpy import numpy as dnp
from .types import Array, ArrayLike
from .utils import grad, jvp


def hvp(
    f: Callable, 
    argnums: int = 0, 
    *args, **kwargs
) -> Callable:

    """Forward-over-reverse Hessian vector product transform for scalar-output functions.

    This does not behave exactly like other functional transforms. The 
    resulting function takes as its first argument a vector, and the 
    following argments are the original function's positional and 
    keyword arguments.

    Additional aguments are passed to `jax.grad()`.

    Parameters
    ----------
    f : Callable
        Function with scalar-output to transform.
    argnums : int, optional
        Which argument number to take the second derivative for. Default: 0.
    
    Returns
    -------
    Callable
        Function which takes a vector `v` as its first argument, and then `f`'s 
        original arguments. Returns the vector product between the Hessian
        of `f` and `v`.

    Examples
    --------
    >>> from duvida.config import config
    >>> config.set_backend("jax", precision="double")
    >>> from duvida.utils import grad, hessian
    >>> import duvida.numpy as dnp 
    >>> f = lambda x: dnp.sum(x ** 3. + x ** 2. + 4.)
    >>> a = dnp.array([1., 2.])
    >>> f(a)
    Array(22., dtype=float64)
    >>> hvp(f)(dnp.ones_like(a), a) == hessian(f)(a) @ dnp.ones_like(a)
    Array([ True,  True], dtype=bool)
    >>> g = lambda x, y: dnp.sum(x ** 2. + x ** 2. + 4. + y ** 3.)
    >>> b = dnp.array([3., 4.])
    >>> hvp(g)(dnp.ones_like(a), a, b) == hessian(g)(a, b) @ dnp.ones_like(a)
    Array([ True,  True], dtype=bool)
    >>> hvp(g, argnums=1)(dnp.ones_like(a), a, b) == hessian(g, argnums=1)(a, b) @ dnp.ones_like(a)
    Array([ True,  True], dtype=bool)

    """

    grad_fn = partial(grad, *args, **kwargs)

    def _hvp(v: ArrayLike, *f_args, **f_kwargs) -> Array:
        d_args = dnp.asarray(f_args[argnums])
        pre_d_args = [arg for i, arg in enumerate(f_args) if i < argnums]
        post_d_args = [arg for i, arg in enumerate(f_args) if i > argnums]
        
        def _jacobian(f_d0_args):
            return f(
                *pre_d_args, 
                f_d0_args, 
                *post_d_args, 
                **f_kwargs,
            )

        return jvp(grad_fn(_jacobian), (d_args,), (v,))[1]

    return _hvp



"""Generic utilities for JAX and PyTorch."""

from typing import Callable, Iterable, Union
from functools import partial

from carabiner import print_err

from .config import config
from .types import Array, ArrayLike

__backend__ = config.backend

if config.backend == 'jax':
    from jax import jit, jvp, grad, hessian, random, vmap
    from jax.flatten_util import ravel_pytree

    def random_normal(seed: int, device=None) -> Callable:
        """Generate a sample from the Normal distribution.

        Examples
        ========
        >>> gen = random_normal(seed=0)
        >>> a, b = gen((3,)), gen((3,))
        >>> (a == b).all()
        Array(True, dtype=bool)

        """
        key = random.key(seed)
        def _normal(shape, *args, **kwargs):
            return random.normal(
                key, 
                shape=shape, 
                *args, **kwargs
            )
        return _normal

elif config.backend == 'torch':
    from functools import wraps
    from torch import concat as concatenate, compile, normal, zeros, Generator, split
    from torch.random import manual_seed
    from torch.func import jvp, grad, hessian, vmap as vmap_torch
    from torch.utils._pytree import tree_flatten, tree_unflatten
    from torch._dynamo import config as dynamo_config
    dynamo_config.suppress_errors = True
    dynamo_config.capture_scalar_outputs = True

    _COMPILE_WARNINGS = set()

    def vmap(
        f: Callable, 
        in_axes: Union[int, Iterable[int]] = 0, 
        out_axes: Union[int, Iterable[int]] = 0, 
        *args, **kwargs
    ) -> Callable:
        """Vectorizes function over axis of its arguments.

        Examples
        ========
        >>> import duvida.numpy as dnp
        >>> double = lambda z: z * 2
        >>> vmap(double)(dnp.array([1., 2., 3.])).tolist()
        [2.0, 4.0, 6.0]

        """
        return vmap_torch(
            f, 
            in_dims=in_axes, 
            out_dims=out_axes, 
            randomness='different',
            *args, **kwargs
        )
        

    def jit(fn):
        try:
            compiled = compile(
                fn,
                fullgraph=True,
                mode="max-autotune",
            )
        except RuntimeError as e:
            warning = f"[torch.compile] Compiling `{fn}` failed; running eagerly"
            if warning not in _COMPILE_WARNINGS:
                _COMPILE_WARNINGS.add(warning)
                print_err(warning + "\n" + str(e))
            return fn

        @wraps(fn)
        def wrapped(*args, **kwargs):
            try:
                return compiled(*args, **kwargs)
            except Exception as e:
                warning = f"[torch.compile] Running compiled `{fn}` failed; falling back"
                if warning not in _COMPILE_WARNINGS:
                    _COMPILE_WARNINGS.add(warning)
                    print_err(warning + "\n" + str(e))
                return fn(*args, **kwargs)
        return wrapped


    def random_normal(seed: int, device='cpu') -> Callable:
        """Generate a sample from the Normal distribution.

        Examples
        ========
        >>> gen = random_normal(seed=0)
        >>> a, b = gen((3,)), gen((3,))
        >>> a, b
        [1], [2]
        >>> (a == b).all()
        tensor(True)

        """
        generator = Generator(device=device)
        generator.manual_seed(seed)

        def _normal(shape, *args, **kwargs):
            return normal(
                generator=generator, 
                mean=zeros(shape, device=device), 
                std=1.,
                *args, **kwargs
            )

        return _normal

    def ravel_pytree(params):
        """Torch pytree flattener.

        Returns
        -------
        flat : 1-D ndarray / Tensor
        unflatten : Callable[[1-D ndarray], original-shaped pytree]

        Examples
        --------
        >>> import duvida.numpy as dnp
        >>> flat, unravel = ravel_pytree([dnp.ones((2,)), dnp.zeros((3,))])
        >>> flat.shape
        (5,)
        >>> (unravel(flat)[0] == 1.).all()
        True
        """
        leaves, spec = tree_flatten(params)  # 
        sizes = [p.numel() for p in leaves]
        flat = concatenate([p.flatten() for p in leaves])

        def unravel(vec):
            chunks = split(vec, sizes)
            rebuilt = [
                chunk.reshape_as(p) 
                for chunk, p in zip(chunks, leaves)
            ]
            return tree_unflatten(spec, rebuilt)

        return flat, unravel
else:
    raise ValueError(f"Invalid backed: {config.backend}")
    

def reciprocal(f: Callable[[ArrayLike], Array]) -> Callable[[ArrayLike], Array]:

    """Convert a function returning a numeric value to return its reciprocal instead.
    
    Parameters
    ----------
    f : Callable
        Function to convert.
    
    Returns
    -------
    Callable
        Function returning the reciprocal of `f`.

    Examples
    --------
    >>> f = lambda x: x + 1.
    >>> f(3.)
    4.0
    >>> reciprocal(f)(1.)
    0.5

    """

    def _inverted(*args, **kwargs):
        return 1. / f(*args, **kwargs)

    return _inverted


def get_eps():

    """Get machine precision.
    
    """
    
    if config.backend == 'jax':
        from jax.numpy import asarray, finfo

        def converter(x): 
            return asarray(x)

    else:
        from torch import as_tensor, finfo

        def converter(x): 
            return as_tensor(x)
            
    return converter(finfo(converter(1.).dtype).eps)



"""Backend configuration."""

from typing import Optional
from dataclasses import dataclass, field
import importlib
import logging
import os

from carabiner import print_err

_BACKEND_FLAG: str = "DUVIDA_BACKEND"
_PRECISION_FLAG: str = "DUVIDA_PRECISION"

log = logging.getLogger("duvida.config")


@dataclass
class Config:
    backend: Optional[str] = field(default=None)
    precision: Optional[str] = field(default=None)

    def __post_init__(self):
        self.backend = self.backend or os.getenv(_BACKEND_FLAG, "jax")
        self.precision = self.precision or os.getenv(_PRECISION_FLAG, "double")
        self.backend_installed: bool = self._backend_installed()
        # self.set_backend()

    def set_backend(
        self, 
        backend: Optional[str] = None, 
        precision: Optional[str] = None
    ) -> None:
        """Set the autodiff backend for duvida.

        Examples
        --------
        >>> config.set_backend("jax", precision="double"); config
        Config(backend='jax', precision='double')
        >>> config.set_backend("torch", precision="float"); config
        Config(backend='torch', precision='float')
        
        """
        if backend is not None:
            self.backend = backend
        if precision is not None:
            self.precision = precision
        self._import_backend()
        self._set_precision()
        return None

    @staticmethod
    def _set_precision_jax(precision: str) -> None:
        from jax import config as jax_config
        if precision == 'double':
            return jax_config.update('jax_enable_x64', True)
        elif precision == 'float':
            return jax_config.update('jax_enable_x64', False)
        elif precision == 'half':
            # TODO: Find implementation of half precision in JAX. JMP?
            raise NotImplementedError("Half precision not available on JAX.")
        else:
            raise ValueError(f"Precision '{precision}' not valid.")
        log.debug(f"JAX precision set to {precision}")

    @staticmethod
    def _set_precision_torch(precision: str) -> None:
        from torch import set_default_dtype  # , set_default_device
        # set_default_device('cuda' if torch.cuda.is_available() else 'cpu')
        if precision == 'double':
            from torch import float64
            return set_default_dtype(float64)
        elif precision == 'float':
            from torch import float32
            return set_default_dtype(float32)
        elif precision == 'half':
            from torch import float16
            return set_default_dtype(float16)
        else:
            raise ValueError(f"Precision '{precision}' not valid.")
        log.debug(f"torch precision set to {precision}")

    def _set_precision(self) -> None:
        if self.backend_installed:
            if self.backend == 'jax':
                precision_setter = self._set_precision_jax
            elif self.backend == 'torch':
                precision_setter = self._set_precision_torch
        else:
            raise AttributeError(f"Backend '{self.backend}' is not installed.")
        precision_setter(self.precision)
        os.environ[_PRECISION_FLAG] = self.precision
        return None

    def _backend_installed(self) -> bool:
        return importlib.util.find_spec(self.backend) is not None

    def _import_backend(self) -> None:
        if self.backend in ('jax', 'torch'):
            importable = self._backend_installed()
        else:
            raise AttributeError(f"Backend '{self.backend}' is not supported.")

        if importable:
            self.backend_installed = importable
        else:
            raise ImportError(
                f"""
                Backend '{self.backend}' could not be imported.
                Try reinstalling duvida with:
                    pip install duvida[{self.backend}]
                or install backend with:
                    pip install {self.backend}
                """
            )
        os.environ[_BACKEND_FLAG] = self.backend
        return None


config = Config()
try:
    config.set_backend('jax', precision='float')
except ImportError:
    print_err("JAX not installed, trying to fall back to torch...")
    try:
        config.set_backend('torch', precision='float')
    except ImportError as e:
        raise e
    else:
        print_err("Duvida using torch backend.")


"""JAX transforms for Hessians diagonals and their approximations."""

from typing import Any, Callable, Dict, Optional, Tuple, Union

from functools import partial

from .hvp import hvp
from .types import Approximator, Array
from .utils import grad, random_normal, vmap, get_eps
from .numpy import numpy as dnp

_EPS = get_eps() ** .5
_DEFAULT_APPROXIMATOR = 'exact_diagonal'


def get_approximators(
    key: Optional[Union[str, Callable]] = None,
    *args, **kwargs
) -> Union[Tuple[str], Dict[str, Approximator]]:

    """Return a list of available Hessian diagonal approximators, or one of the
    named approximator functions.

    Parameters
    ----------
    key : str, optional
        If provided, return the function with the name `key`.

    Returns
    -------
    Union[Tuple[str], Dict[str, Approximator]]
        If key is provided, returns the named approximator function. Otherwise
        returns a tuple of named approximators.

    Raises
    ------
    NotImplementedError
        When `key` is provided but is not in the list of approximators.

    """

    APPROXIMATORS = {
        'squared_jacobian': squared_jacobian,
        'exact_diagonal': exact_diagonal,
        'bekas': bekas,
        'rough_finite_difference': rough_finite_difference,
    }

    if key is None:
        return tuple(APPROXIMATORS)
    elif isinstance(key, str):
        try:
            return partial(APPROXIMATORS[key], *args, **kwargs)
        except KeyError:
            NotImplementedError(
                f"Approximator called '{key}' is not implemented. Choose from {', '.join(sorted(APPROXIMATORS))}"
            )
    elif isinstance(key, Callable):
        return key
    else:
        raise ValueError(f"Key must be a str or Callable, but was actually '{type(key)}'.")


def squared_jacobian(
    f: Callable[[Any], Array], 
    *args, **kwargs
) -> Callable[[Any], Array]:

    """Functional transform to the square of the Jacobian of a scalar-output function.
    
    Additional aguments are passed to `jax.grad()`. Set `argnums` to
    differentiate `f` with respect to parameters other than the first.

    Parameters
    ----------
    f : Callable
        Function with scalar-output to transform.
    
    Returns
    -------
    Callable
        Function which takes `f`'s original arguments. Returns the 
        square of the gradient of `f` with respect to its first parameter.

    Examples
    --------
    >>> from duvida.utils import hessian
    >>> import duvida.numpy as dnp  
    >>> f = lambda x: dnp.sum(x ** 3. + x ** 2. + 4.)
    >>> a = dnp.array([1., 2.])
    >>> f(a)
    Array(22., dtype=float64)
    >>> dnp.diag(hessian(f)(a))
    Array([ 8., 14.], dtype=float64)
    >>> squared_jacobian(f)(a) == dnp.square(grad(f)(a))
    Array([ True,  True], dtype=bool)
    >>> squared_jacobian(f)(a)  # Not very accurate!
    Array([ 25., 256.], dtype=float64)
    
    """
    _jacobian = grad(f, *args, **kwargs)

    def _sq_jacobian(*args, **kwargs) -> Array:
        return dnp.square(_jacobian(*args, **kwargs))

    return _sq_jacobian


def exact_diagonal(
    f: Callable[[Any], Array], 
    argnums: int = 0,
    device: str = 'cpu',
    *args, **kwargs
) -> Callable[[Any], Array]:

    """Functional transform to the Hessian diagonal of a scalar-output function.
    
    Additional aguments are passed to `hvp()`. Set `argnums` to
    differentiate `f` with respect to parameters other than the first.

    Parameters
    ----------
    f : Callable
        Function with scalar-output to transform.
    argnums : int, optional
        Which argument number to take the second derivative for. Default: 0.
    
    Returns
    -------
    Callable
        Function which takes `f`'s original arguments. Returns the 
        diagonal of the Hessian of `f`. 

    Examples
    --------
    >>> from duvida.utils import hessian
    >>> import duvida.numpy as dnp  
    >>> f = lambda x: dnp.sum(x ** 3. + x ** 2. + 4.)
    >>> a = dnp.array([1., 2.])
    >>> exact_diagonal(f)(a) == dnp.diag(hessian(f)(a))
    Array([ True,  True], dtype=bool)
    >>> g = lambda x: dnp.sum(dnp.sum(x) ** 3. + x ** 2. + 4.)
    >>> exact_diagonal(g)(a) == dnp.diag(hessian(g)(a))
    Array([ True,  True], dtype=bool)
    
    """
    hvp_f = hvp(f, argnums=argnums, *args, **kwargs)

    def get_hessian_element(
        i: int, 
        size: int, 
        *args, **kwargs
    ) -> float:
        unit_vec = dnp.one_hot(i, size, device=device)
        return dnp.take(hvp_f(unit_vec, *args, **kwargs), i)

    def _hessian_diagonal(*args, **kwargs) -> Array:
        d_args = dnp.asarray(args[argnums])
        d_args_size = dnp.get_array_size(d_args)
        v_hvp_f = vmap(
            get_hessian_element, 
            in_axes=(0, None) + (None, ) * len(args)
        )
        idx = dnp.arange(d_args_size, device=device)
        return v_hvp_f(idx, d_args_size, *args, **kwargs) 

    return _hessian_diagonal


def bekas(
    f: Callable[[Any], Array], 
    n: int = 1, 
    seed: int = 0,
    argnums: int = 0,
    device: str = 'cpu',
    *args, **kwargs
) -> Callable[[Any], Array]:

    """Functional transform to the Bekas estimator of the Hessian diagonal of a scalar-output function.
    
    Additional aguments are passed to `hvp()`. Set `argnums` to
    differentiate `f` with respect to parameters other than the first.

    Bekas estimator is described at https://doi.org/10.1016/j.apnum.2007.01.003.

    Parameters
    ----------
    f : Callable
        Function with scalar-output to transform.
    n : int, optional
        Number of Gaussian random samples to take. Default: 1.
    seed : int, optional
        Controls the random behavior. Set a different seed to get a 
        different result. Default: 0.
    argnums : int, optional
        Which argument number to take the second derivative for. Default: 0.
    device : str, optional
        For PyTorch backend. Which device to use: "cpu" or "cuda". Default: "cpu".
    
    Returns
    -------
    Callable
        Function which takes `f`'s original arguments. Returns the 
        Bekas approximation of the diagonal of the Hessian of `f`. 

    Examples
    --------
    >>> from duvida.utils import hessian
    >>> import duvida.numpy as dnp  
    >>> f = lambda x: dnp.sum(x ** 3. + x ** 2. + 4.)
    >>> a = dnp.array([1., 2.])
    >>> f(a)
    Array(22., dtype=float64)
    >>> dnp.diag(hessian(f)(a))
    Array([ 8., 14.], dtype=float64)
    >>> bekas(f)(a)  # Apparently very accurate
    Array([ 8., 14.], dtype=float64)
    >>> dnp.allclose(bekas(f)(a), dnp.diag(hessian(f)(a)))
    Array(True, dtype=bool)
    >>> bekas(f, n=100_000, seed=1)(a)  # Increase samples for better accuracy
    Array([ 8., 14.], dtype=float64)
    >>> bekas(f, seed=1)(a) == bekas(f, seed=0)(a)  # Change the seed to alter the outcome
    Array([ True, False], dtype=bool)
    >>> g = lambda x: dnp.sum(dnp.sum(x) ** 3. + x ** 2. + 4.)
    >>> dnp.diag(hessian(g)(a))
    Array([38., 38.], dtype=float64)
    >>> bekas(g, n=1000, seed=0)(a)  # Less accurate when parameters interact
    Array([39.50905509, 39.81868951], dtype=float64)
    >>> bekas(g, n=1000, seed=1)(a)  # Change the seed to alter the outcome
    Array([37.25647008, 37.29415901], dtype=float64)
    
    """
    random_normal_fn = random_normal(seed, device=device)
    hvp_f = hvp(f, argnums=argnums, *args, **kwargs)

    def _approx_hessian_diagonal(*args, **kwargs) -> Array:
        d_args = args[argnums]
        d_args_size = dnp.get_array_size(d_args)
        v_hvp_f = vmap(
            hvp_f, 
            in_axes=(1, ) + (None, ) * len(args), 
            out_axes=1,
        )
        v = random_normal_fn(shape=(d_args_size, n))  # p, n  # TODO: Don't instantiate all at once - risk of memory blow-up
        samples = v * v_hvp_f(v, *args, **kwargs)   # p, n
        return dnp.sum(samples, axis=-1) / dnp.sum(dnp.square(v), axis=-1)

    return _approx_hessian_diagonal


def rough_finite_difference(
    f: Callable[[Any], Array], 
    argnums: int = 0, 
    eps: float = _EPS,
    *args, **kwargs
) -> Callable[[Any], Array]:
    
    """Functional transform to a rough Hessian diagonal of a scalar-output function.

    This uses the usual finite-difference algorithm from `scipy.optimize.fprime()`
    to calculate the Jacobian of the autodiff gradients of `f`, but changes 
    all values at once, instead of one-by-one. 
    
    This will work best when parameters don't have strong interactions with
    each other becuase it's essentially a column sum of the Hessian matrix.
    
    Additional aguments are passed to `grad()`. Set `argnums` to
    differentiate `f` with respect to parameters other than the first.

    Parameters
    ----------
    f : Callable
        Function with scalar-output to transform.
    argnums : int, optional
        Which argument number to take the second derivative for. Default: 0.
    eps : float, optional
        The small number to add to prevent instability.
    
    Returns
    -------
    Callable
        Function which takes `f`'s original arguments. Returns the 
        rough finite-difference approximation of the diagonal of the 
        Hessian of `f`. 

    Examples
    --------
    >>> from duvida.utils import hessian
    >>> import duvida.numpy as dnp 
    >>> f = lambda x: dnp.sum(x ** 3. + x ** 2. + 4.)
    >>> a = dnp.array([1., 2.])
    >>> f(a)
    Array(22., dtype=float64)
    >>> dnp.diag(hessian(f)(a))
    Array([ 8., 14.], dtype=float64)
    >>> rough_finite_difference(f)(a)  # Relatively accurate
    Array([ 8.0010358, 14.0010358], dtype=float64)
    >>> rough_finite_difference(f, eps=.01)(a)
    Array([ 8.03, 14.03], dtype=float64)
    >>> g = lambda x: dnp.sum(dnp.sum(x) ** 3. + x ** 2. + 4.)
    >>> dnp.diag(hessian(g)(a))
    Array([38., 38.], dtype=float64)
    >>> rough_finite_difference(g)(a)  # Less accurate when parameters interact
    Array([74.00828641, 74.00828641], dtype=float64)
    
    """
    _jacobian = grad(f, argnums=argnums, *args, **kwargs)

    def _approx_hessian_diagonal(*args, **kwargs) -> Array:
        d_args = [arg + eps if i == argnums else arg for i, arg in enumerate(args)]
        return (_jacobian(*d_args, **kwargs) - _jacobian(*args, **kwargs)) / eps

    return _approx_hessian_diagonal


if _DEFAULT_APPROXIMATOR not in get_approximators():  # Should never happen
    raise KeyError(f"Default approximator {_DEFAULT_APPROXIMATOR} not in approximator dictionary!")  



"""Generic types for JAX and PyTorch"""

from typing import TYPE_CHECKING, Any, Callable, Iterable, Union

from .config import config

if TYPE_CHECKING:  
    if config.backend == 'jax':
        from jax import Array
        from jax.typing import ArrayLike
    else:
        if config.precision == 'double':
            from torch import DoubleTensor as Array
        elif config.precision == 'float':
            from torch import FloatTensor as Array
        elif config.precision == 'half':
            from torch import HalfTensor as Array
        else:
            from torch import FloatTensor as Array
        ArrayLike = Union[Array, float]
else:
    Array, ArrayLike = Any, Any

Approximator = Callable[[Callable], Callable]
LossFunction = Callable[[ArrayLike, ArrayLike], float]
StatelessModel = Callable[[Iterable[ArrayLike], ArrayLike], Array]



from importlib.metadata import version

appname = "duvida"
__version__ = version(appname)
__author__ = "Eachan Johnson"

from .hessians import (
    bekas,
    exact_diagonal,
    get_approximators, 
    rough_finite_difference,
    squared_jacobian
)
from .hvp import hvp
from .information import (
    doubtscore,
    fisher_score,
    fisher_information_diagonal,
    parameter_gradient,
    parameter_hessian_diagonal,
    information_sensitivity
)
from .utils import *



"""Implementation of information sensitivity metrics for stateless model frameworks."""

from typing import Callable, Union
from functools import partial

from .hessians import _DEFAULT_APPROXIMATOR, get_approximators
from .numpy import numpy as dnp
from .types import Array, ArrayLike, LossFunction, StatelessModel
from .utils import reciprocal, grad, jit, vmap


def parameter_gradient(model: StatelessModel) -> Callable[[ArrayLike, ArrayLike], Array]:

    """Gradient of a scalar function with respect to its parameters for each 
    point in x.
    
    The input function must take x as its first argument, and unpacked parameters
    as its other arguments. The output has x as its second argument and an 
    iterable of parameters as its first arguments.

    Parameters
    ----------
    model : Callable
        Function to calcuate the gradient of with respect to the parameters
        in its second argument.

    Returns
    -------
    Callable
        Function returning parameter gradient of `f`.

    Examples
    --------    
    >>> import duvida.numpy as dnp
    >>> f = lambda x, p1, p2: x ** p1 + p2
    >>> x = dnp.array([1., 2.])
    >>> p = dnp.array([2., 1.])
    >>> f(x, *p)
    Array([2., 5.], dtype=float64)
    >>> parameter_gradient(f)(p, x)  # doctest: +NORMALIZE_WHITESPACE
    Array([[0.        , 1.        ],
           [2.77258872, 1.        ]], dtype=float64)

    """
    
    @partial(vmap, in_axes=(None, 0))
    @grad
    def _parameter_gradient(
        params: ArrayLike, 
        x: ArrayLike
    ) -> Array:
        return dnp.sum(model(x, *params))

    return _parameter_gradient


def parameter_hessian_diagonal(
    f: StatelessModel, 
    approximator: str = _DEFAULT_APPROXIMATOR, 
    *args, **kwargs
) -> Callable[[ArrayLike, ArrayLike], Array]:

    """Diagonal of the Hessian of a scalar function with respect to its parameters for each 
    point in x.
    
    The input function must take an iterable of parameters as its second
    argument, and take x values as its first argument. The output has 
    x as its first argument and unpacked parameters as its other arguments.

    Additional arguments are passed to `approximator` function.

    Parameters
    ----------
    f : Callable
        Function to calcuate the gradient of with respect to the parameters
        in its second argument.
    approximator : str, optional
        Named approximator for diagonal of Hessian. Default: use exact Hessian diagonal.
    
    Returns
    -------
    Callable
        Function returning diagonal of the Hessian of `f`.

    Examples
    --------    
    >>> import duvida.numpy as dnp 
    >>> f = lambda x, p1, p2: x ** p1 + p2
    >>> x = dnp.array([1., 2.])
    >>> p = dnp.array([2., 1.])
    >>> f(x, *p)
    Array([2., 5.], dtype=float64)
    >>> parameter_hessian_diagonal(f)(p, x)  # doctest: +NORMALIZE_WHITESPACE
    Array([[0.        , 0.        ],
           [1.92181206, 0.        ]], dtype=float64)
    >>> parameter_hessian_diagonal(f, approximator='squared_jacobian')(p, x)  # doctest: +NORMALIZE_WHITESPACE
    Array([[0.        , 1.        ],
           [7.68724822, 1.        ]], dtype=float64)
    >>> parameter_hessian_diagonal(f, approximator='rough_finite_difference')(p, x)  # doctest: +NORMALIZE_WHITESPACE
    Array([[0.        , 0.        ],
           [1.92204204, 0.        ]], dtype=float64)
    >>> parameter_hessian_diagonal(f, approximator='bekas', n=3, seed=0)(p, x)  # doctest: +NORMALIZE_WHITESPACE
    Array([[0.        , 0.        ],
           [1.92181206, 0.        ]], dtype=float64)

    """
    
    @partial(vmap, in_axes=(None, 0))
    @get_approximators(approximator, *args, **kwargs)
    def _scalar_f(
        params: ArrayLike, 
        x: ArrayLike
    ) -> Array:
        return dnp.sum(f(x, *params))      

    return _scalar_f


def parameter_gradient_unrolled(
    model: StatelessModel
) -> Callable[[ArrayLike, ArrayLike], Array]:

    @grad
    def _f0(
        params: ArrayLike, 
        x: ArrayLike
    ) -> Array:
        return dnp.sum(model(x, *params))

    def _f(
        params: ArrayLike, 
        x: ArrayLike
    ) -> Array:
        return [_f0(params, [d]) for d in x]

    return _f


def parameter_hessian_diagonal_unrolled(
    model: StatelessModel, 
    approximator: str = _DEFAULT_APPROXIMATOR, 
    *args, **kwargs
) -> Callable[[ArrayLike, ArrayLike], Array]:

    @get_approximators(approximator, *args, **kwargs)
    def _f0(
        params: ArrayLike, 
        x: ArrayLike
    ) -> Array:
        return dnp.sum(model(x, *params))

    def _f(
        params: ArrayLike, 
        x: ArrayLike
    ) -> Array:
        return [_f0(params, [d]) for d in x]

    return _f


def _model_loss(
    model: StatelessModel, 
    loss: LossFunction
) -> Callable[[ArrayLike, ArrayLike, ArrayLike], float]:
    
    def _loss(
        params: ArrayLike, 
        x_true: ArrayLike, 
        y_true: ArrayLike
    ) -> Array:
        return loss(model(x_true, *params), y_true)

    return _loss


def fisher_score(
    model: StatelessModel, 
    loss: LossFunction
) -> Callable[[ArrayLike, ArrayLike, ArrayLike], Array]:
    
    """Returns a function for the Fisher score of a model's loss with respect to
    its parameters.
    
    The input function model takes an iterable of parameters as its second
    argument, and take x values as its first argument. The output has 
    parameter iterable as its first argument, observed x-values as its second 
    argument, and observed y-values as its third argument.

    Parameters
    ----------
    model : Callable
        Function to convert.
    loss : Callable
        Function taking predictions and observations, and giving a loss value.
    
    Returns
    -------
    Callable
        Function returning the Fisher score of `model`'s `loss` with respect to
        its parameters.

    Examples
    --------   
    >>> import duvida.numpy as dnp  
    >>> model = lambda x, p1, p2: x ** p1 + p2
    >>> mse_fn = lambda ypred, ytrue: dnp.sum(dnp.square(ypred - ytrue))
    >>> x = dnp.array([1., 2.])
    >>> p = dnp.array([0., 2.])
    >>> model(x, *p)
    Array([3., 3.], dtype=float64)
    >>> fisher_score(model, mse_fn)(p, x, model(x, *p) + .1)
    Array([-0.13862944, -0.4       ], dtype=float64)

    """

    return grad(_model_loss(model, loss))


def fisher_information_diagonal(
    model: StatelessModel, 
    loss: LossFunction, 
    approximator: Union[str, Callable] = _DEFAULT_APPROXIMATOR,
    *args, **kwargs
) -> Callable[[ArrayLike, ArrayLike, ArrayLike, ArrayLike], Array]:

    """Returns a function for the diagonal of the Fisher information of 
    a model's loss with respect to its parameters.
    
    The input function model takes an iterable of parameters as its second
    argument, and take x values as its first argument. The output has 
    parameter iterable as its first argument, observed x-values as its second 
    argument, and observed y-values as its third argument.

    Additional arguments are passed to `approximator` function.

    Parameters
    ----------
    model : Callable
        Function to convert.
    loss : Callable
        Function taking predictions and observations, and giving a loss value.
    approximator : str, optional
        Named approximator for diagonal of Hessian. Default: use exact Hessian diagonal.
    
    Returns
    -------
    Callable
        Function returning the Fisher score of `model`'s `loss` with respect to
        its parameters.

    Examples
    --------    
    >>> import duvida.numpy as dnp  
    >>> model = lambda x, p1, p2: x ** p1 + p2
    >>> mse_fn = lambda ypred, ytrue: dnp.sum(dnp.square(ypred - ytrue))
    >>> x = dnp.array([1., 2.])
    >>> p = dnp.array([0., 2.])
    >>> model(x, *p)
    Array([3., 3.], dtype=float64)
    >>> fisher_information_diagonal(model, mse_fn)(p, x, model(x, *p) + .1)
    Array([0.86481543, 4.        ], dtype=float64)

    """

    return get_approximators(approximator)(
        _model_loss(model, loss), 
        *args, **kwargs,
    )


def doubtscore(
    model: StatelessModel, 
    loss: LossFunction, 
    use_reciprocal: bool = False
) -> Callable[[ArrayLike, ArrayLike, ArrayLike, ArrayLike], Array]:

    """Given a model and loss function, returns a function for the 
    doubtscore of points in the domain.
    
    The input function model takes an iterable of parameters as its second
    argument, and takes domain x values as its first argument. 
    
    The output function has a parameter iterable as its first argument, 
    domain x values as its second argument, observed x-values as its third 
    argument, and observed y-values as its fourth argument.

    Parameters
    ----------
    model : Callable
        Function to convert.
    loss : Callable
        Function taking predictions and observations, and giving a loss value.
    reciprocal : bool, optional
        Whether the returned function should instead give the reciprocal
        doubtscore. Default: `False`.
    
    Returns
    -------
    Callable
        Function returning the doubtscore for values of the domain and 
        observed data.

    Examples
    --------   
    >>> import duvida.numpy as dnp   
    >>> model = lambda x, p1, p2: x ** p1 + p2
    >>> mse_fn = lambda ypred, ytrue: dnp.sum(dnp.square(ypred - ytrue))
    >>> x = dnp.array([1., 2.])
    >>> p = dnp.array([0., 2.])
    >>> model(x, *p)
    Array([3., 3.], dtype=float64)
    >>> doubtscore(model, mse_fn)(p, x + .1, x, model(x, *p) + .1)
    Array([[-1.45450818, -0.4       ],
           [-0.1868479 , -0.4       ]], dtype=float64)

    """

    param_grad_fn = parameter_gradient(model)
    fisher_score_fn = fisher_score(model, loss)

    def _doubtscore(params: ArrayLike, x: ArrayLike, 
                    x_true: ArrayLike, y_true: ArrayLike) -> Array:
        return (
            fisher_score_fn(params, x_true, y_true) 
            / param_grad_fn(params, x)
        )

    if use_reciprocal:
        return jit(reciprocal(_doubtscore))
    else:
        return jit(_doubtscore)
    

def _information_sensitivity_term1(
    model: StatelessModel, 
    loss: LossFunction, 
    approximator: Union[str, Callable] = _DEFAULT_APPROXIMATOR,
    *args, **kwargs
) -> Callable[[ArrayLike, ArrayLike, ArrayLike, ArrayLike], Array]:
    param_grad_fn = parameter_gradient(model)
    fisher_info_fn = fisher_information_diagonal(model, loss, approximator, 
                                                 *args, **kwargs)

    def _term1(params: ArrayLike, x: ArrayLike, 
               x_true: ArrayLike, y_true: ArrayLike) -> Array:
        return fisher_info_fn(params, x_true, y_true) / param_grad_fn(params, x)

    return _term1


def _information_sensitivity_term2(
    model: StatelessModel, 
    loss: LossFunction, 
    approximator: Union[str, Callable] = _DEFAULT_APPROXIMATOR,
    *args, **kwargs
) -> Callable[[ArrayLike, ArrayLike, ArrayLike, ArrayLike], Array]:
    fisher_score_fn = fisher_score(model, loss)
    param_grad_fn = parameter_gradient(model)
    param_hessian_fn = parameter_hessian_diagonal(model, approximator, 
                                                  *args, **kwargs)

    def _term2(
        params: ArrayLike, 
        x: ArrayLike, 
        x_true: ArrayLike, 
        y_true: ArrayLike
    ) -> Array:
        return (fisher_score_fn(params, x_true, y_true) * param_hessian_fn(params, x)
                / dnp.square(param_grad_fn(params, x)))

    return _term2


def information_sensitivity(
    model: StatelessModel, 
    loss: LossFunction, 
    approximator: Union[str, Callable] = _DEFAULT_APPROXIMATOR,
    use_reciprocal: bool = False,
    *args, **kwargs
) -> Callable[[ArrayLike, ArrayLike, ArrayLike, ArrayLike], Array]:

    """Given a model and loss function, returns a function for the 
    inforamtion sensitivity of points in the domain.
    
    The input function model takes an iterable of parameters as its second
    argument, and takes domain x values as its first argument. 
    
    The output function has a parameter iterable as its first argument, 
    domain x values as its second argument, observed x-values as its third 
    argument, and observed y-values as its fourth argument.

    Additional arguments are passed to `approximator` function.

    Parameters
    ----------
    model : Callable
        Function to convert.
    loss : Callable
        Function taking predictions and observations, and giving a loss value.
    pproximator : str, optional
        Named approximator for diagonal of Hessian. Default: use exact Hessian diagonal.
    reciprocal : bool, optional
        Whether the returned function should instead give the reciprocal
        information sensitivity. Default: `False`.
    
    Returns
    -------
    Callable
        Function returning the information sensitivity for values of the domain and 
        observed data.

    Examples
    --------    
    >>> from duvida.config import config
    >>> model = lambda x, p1, p2: x ** p1 + p2
    >>> mse_fn = lambda ypred, ytrue: dnp.sum(dnp.square(ypred - ytrue))
    >>> x = dnp.array([1., 2.])
    >>> p = dnp.array([0., 2.])
    >>> model(x, *p)
    Array([3., 3.], dtype=float64)
    >>> information_sensitivity(model, mse_fn)(p, x + .1, x, model(x, *p) + .1)
    Array([[9.21232363, 4.        ],
           [1.3042473 , 4.        ]], dtype=float64)

    """

    term1_fn, term2_fn = (
        f(model, loss, approximator, *args, **kwargs) 
        for f in (
            _information_sensitivity_term1, 
            _information_sensitivity_term2,
        )
    )

    def _information_sensitivity(
        params: ArrayLike, 
        x: ArrayLike, 
        x_true: ArrayLike, 
        y_true: ArrayLike
    ) -> Array:
        return term1_fn(params, x, x_true, y_true) - term2_fn(params, x, x_true, y_true)

    if use_reciprocal:
        return jit(reciprocal(_information_sensitivity))
    else:
        return jit(_information_sensitivity)



"""Very rough backend-agnostic NumPy API."""

from ..config import config
from ..types import Array, ArrayLike

__backend__ = config.backend

if config.backend == 'jax':
    from jax.numpy import (
        float64, int64, 
        allclose, 
        arange as jax_arange, 
        asarray, array, concatenate, diag, 
        expand_dims, ones, ones_like, roll, sum, sqrt, square, take, zeros,
        zeros_like
    )
    from jax.nn import one_hot as jax_one_hot

    def arange(start, stop=None, step=None, device: str = "cpu") -> Array:
        return jax_arange(start, stop, step)

    def get_array_size(a: ArrayLike) -> int:
        return asarray(a).size

    def set_array_element(a: ArrayLike, i: int, x: ArrayLike) -> Array:
        return asarray(a).at[i].set(x)

    def unsqueeze(a: ArrayLike, axis: int) -> Array:
        return expand_dims(asarray(a), axis)

    def dtype_like(a: ArrayLike, b: ArrayLike) -> Array:
        return asarray(a).astype(asarray(b).dtype)

    def one_hot(tensor: ArrayLike, num_classes: int = -1, device: str = 'cpu') -> Array:
        return dtype_like(jax_one_hot(tensor, num_classes), 1.)

else:
    from torch import (
        float64, int64, 
        as_tensor as asarray, 
        Tensor as array, 
        arange, 
        allclose, 
        concat as concatenate, 
        diag, numel, 
        ones, ones_like, 
        roll, sum, sqrt, square, take, 
        zeros, zeros_like
    )
    from torch.nn.functional import one_hot as torch_one_hot

    def get_array_size(a: ArrayLike):
        return numel(asarray(a))

    def set_array_element(a: ArrayLike, i: int, x: ArrayLike) -> Array:
        a_copy = a.detach().clone()
        a_copy[i] = x
        return a_copy

    def unsqueeze(a: ArrayLike, axis: int) -> Array:
        return asarray(a).unsqueeze(axis)
    
    def dtype_like(a: ArrayLike, b: ArrayLike) -> Array:
        return asarray(a).to(asarray(b).dtype)

    def one_hot(tensor: ArrayLike, num_classes: int = -1, device: str = 'cpu') -> Array:
        return dtype_like(torch_one_hot(tensor, num_classes), 1.).to(device)
